"""
Handle Asynchronous Telnet Connections.
Single-process, no threads, select-based.

Limitations:
 - No linemode support, character-at-a-time only.
 - No out-of-band / data mark (DM) / sync supported
   (no ^C, ^S, ^Q helpers)

This is a modified version of miniboa retrieved from
svn address http://miniboa.googlecode.com/svn/trunk/miniboa
which is meant for MUD's. This server would not be safe for MUD clients.
"""
#  Copyright 2012 Jeff Quast, whatever Jim's license is; changes from miniboa:
#    character-at-a-time input instead of linemode, encoding option on send,
#    strict rejection of linemode, terminal type detection, environment
#    variable support, GA and SGA, utf-8 safe

#------------------------------------------------------------------------------
#   miniboa/async.py
#   miniboa/telnet.py
#   Copyright 2009 Jim Storch
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain a
#   copy of the License at http://www.apache.org/licenses/LICENSE-2.0
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#------------------------------------------------------------------------------

import warnings
import socket
import array
import time
import logging

# regretable
from x84.bbs.exception import Disconnected

# pylint: disable=C0103
#        Invalid name "logger" for type constant
logger = logging.getLogger()

#--[ Telnet Options ]----------------------------------------------------------
from telnetlib import LINEMODE, NAWS, NEW_ENVIRON, ENCRYPT, AUTHENTICATION
from telnetlib import BINARY, SGA, ECHO, STATUS, TTYPE, TSPEED, LFLOW
from telnetlib import XDISPLOC, IAC, DONT, DO, WONT, WILL, SE, NOP, DM, BRK
from telnetlib import IP, AO, AYT, EC, EL, GA, SB
IS = chr(0)  # Sub-process negotiation IS command
SEND = chr(1)  # Sub-process negotiation SEND command
UNSUPPORTED_WILL = (LINEMODE, LFLOW, TSPEED, ENCRYPT, AUTHENTICATION)


class TelnetServer(object):
    """
    Poll sockets for new connections and sending/receiving data from clients.
    """
    MAX_CONNECTIONS = 1000
    TIME_POLL = 0.05
    LISTEN_BACKLOG = 5
    ## Dictionary of active clients, (file descriptor, TelnetClient,)
    clients = {}
    ## Dictionary of environment variables received by negotiation
    env = {}

    def __init__(self, address_pair, on_naws):
        """
        Create a new Telnet Server.

        Arguments:
           address_pair: tuple of (ip, port) to bind to.
           on_naws: this callable receives a TelnetClient when a client
                    negotiates about window size (resize event).
        """
        (self.address, self.port) = address_pair
        self.on_naws = on_naws

        # bind
        self.server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server_socket.setsockopt(
            socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            self.server_socket.bind(address_pair)
            self.server_socket.listen(self.LISTEN_BACKLOG)
        except socket.error as err:
            logger.error('Unable to bind: %s', err)
            exit(1)

    def client_count(self):
        """
        Returns the number of active connections.
        """
        return len(self.clients)

    def client_list(self):
        """
        Returns a list of connected clients.
        """
        return self.clients.values()

#---[ Telnet Notes ]-----------------------------------------------------------
# (See RFC 854 for more information)
#
# Negotiating a Local Option
# --------------------------
#
# Side A begins with:
#
#    "IAC WILL/WONT XX"   Meaning "I would like to [use|not use] option XX."
#
# Side B replies with either:
#
#    "IAC DO XX"     Meaning "OK, you may use option XX."
#    "IAC DONT XX"   Meaning "No, you cannot use option XX."
#
#
# Negotiating a Remote Option
# ----------------------------
#
# Side A begins with:
#
#    "IAC DO/DONT XX"  Meaning "I would like YOU to [use|not use] option XX."
#
# Side B replies with either:
#
#    "IAC WILL XX"   Meaning "I will begin using option XX"
#    "IAC WONT XX"   Meaning "I will not begin using option XX"
#
#
# The syntax is designed so that if both parties receive simultaneous requests
# for the same option, each will see the other's request as a positive
# acknowledgement of it's own.
#
# If a party receives a request to enter a mode that it is already in, the
# request should not be acknowledged.

## Where you see DE in my comments I mean 'Distant End', e.g. the client.

UNKNOWN = -1

#-----------------------------------------------------------------Telnet Option


class TelnetOption(object):
    """
    Simple class used to track the status of an extended Telnet option.
    Attributes and values:
        local_option: UNKNOWN (default), True, or False.
        remote_option: UNKNOWN (default), True, or False.
        reply_pending: True or Fale.
    """
    # pylint: disable=R0903
    #         Too few public methods (0/2)
    def __init__(self):
        """
        Set attribute defaults on init.
        """
        self.local_option = UNKNOWN     # Local state of an option
        self.remote_option = UNKNOWN    # Remote state of an option
        self.reply_pending = False      # Are we expecting a reply?


def name_option(option):
    """
    Perform introspection of global CONSTANTS for equivalent values,
    and return a string that displays its possible meanings
    """
    values = ';?'.join([k for k, v in globals().iteritems()
                        if option == v and k not in ('SEND', 'IS',)])
    return values if values != '' else str(ord(option))

# import os, inspect
# def debug_option(func):
#    """
#    This function is a decorator that debug prints the 'from' address for
#    callables decorated with this. This helps during telnet negotiation, to
#    understand which function sets or checks local or remote option states.
#    """
#    def wrapper(self, *args):
#        """
#        inner wrapper for debug_option
#        """
#        stack = inspect.stack()
#        logger.debug ('%s:%s %s(%s%s)',
#            os.path.basename(stack[1][1]), stack[1][2],
#            func.__name__, name_option(args[0]),
#            ', %s' % (args[1],) if len(args) == 2 else '')
#        return func(self, *args)
#    return wrapper


#------------------------------------------------------------------------Telnet

class TelnetClient(object):
    """
    Represents a remote Telnet Client, instantiated from TelnetServer.
    """
    # pylint: disable=R0902,R0904
    #         Too many instance attributes
    #         Too many public methods
    BLOCKSIZE_RECV = 64
    SB_MAXLEN = 65534  # maximum length of subnegotiation string, allow
                       # a fairly large one for NEW_ENVIRON negotiation

    def __init__(self, sock, address_pair, on_naws=None):
        """
        Arguments:
            sock: socket
            address_pair: tuple (ip address, port number)
            on_naws: callback for window resizing by client
        """
        self.sock = sock
        self.address_pair = address_pair
        self.on_naws = on_naws
        self.active = True
        self.env = dict([('TERM', 'unknown')])
        self.send_buffer = array.array('c')
        self.recv_buffer = array.array('c')
        self.telnet_sb_buffer = array.array('c')
        self.bytes_received = 0
        self.connect_time = time.time()
        self.last_input_time = time.time()

        ## State variables for interpreting incoming telnet commands
        self.telnet_got_iac = False
        self.telnet_got_cmd = None
        self.telnet_got_sb = False
        self.telnet_opt_dict = {}

        self.ENV_REQUESTED = False
        self.ENV_REPLIED = False

    def get_input(self):
        """
        Get any input bytes received from the DE. The input_ready method
        returns True when bytes are available.
        """
        data = self.recv_buffer.tostring()
        self.recv_buffer = array.array('c')
        return data

    def send_str(self, bstr):
        """
        Buffer bytestring for client.
        """
        self.send_buffer.fromstring(bstr)

    def send_unicode(self, ucs, encoding='utf8'):
        """
        Buffer unicode string, encoded for client as 'encoding'.
        """
        ## Must be escaped 255 (IAC + IAC) to avoid IAC intepretation
        self.send_str(ucs.encode(encoding, 'replace')
                      .replace(chr(255), 2 * chr(255)))

    def deactivate(self):
        """
        Flag client for disconnection.
        """
        if not self.active:
            logger.debug('%s: already deactivated', self.addrport())
            return
        logger.debug('%s: deactivated', self.addrport())
        self.active = False

    def addrport(self):
        """
        Returns IP address and port of DE as string.
        """
        return '%s:%d' % (self.address_pair[0], self.address_pair[1])

    def idle(self):
        """
        Returns time elapsed since DE last sent input.
        """
        return time.time() - self.last_input_time

    def duration(self):
        """
        Returns time elapsed since DE connected.
        """
        return time.time() - self.connect_time

    def request_will_sga(self):
        """
        Request DE to Suppress Go-Ahead.  See RFC 858.
        """
        self._iac_will(SGA)
        self._note_reply_pending(SGA, True)

    def request_will_echo(self):
        """
        Tell the DE that we would like to echo their text.  See RFC 857.
        """
        self._iac_will(ECHO)
        self._note_reply_pending(ECHO, True)

    def request_will_binary(self):
        """
        Tell the DE that we would like to use binary 8-bit (utf8).
        """
        self._iac_will(BINARY)
        self._note_reply_pending(BINARY, True)

    def request_do_binary(self):
        """
        Tell the DE that we would like them to input binary 8-bit (utf8).
        """
        self._iac_do(BINARY)
        self._note_reply_pending(BINARY, True)

    def request_do_sga(self):
        """
        Request to Negotiate SGA.  See ...
        """
        self._iac_do(SGA)
        self._note_reply_pending(SGA, True)

    def request_do_naws(self):
        """
        Request to Negotiate About Window Size.  See RFC 1073.
        """
        self._iac_do(NAWS)
        self._note_reply_pending(NAWS, True)

    def request_do_env(self):
        """
        Request to Negotiate About Window Size.  See RFC 1073.
        """
        self._iac_do(NEW_ENVIRON)
        self._note_reply_pending(NEW_ENVIRON, True)

    def request_env(self):
        """
        Request sub-negotiation NEW_ENVIRON. See RFC 1572.
        """
        if self.ENV_REQUESTED:
            return  # avoid asking twice ..
        rstr = bytes(''.join((IAC, SB, NEW_ENVIRON, SEND, chr(0))))
        rstr += bytes(chr(0).join(
            ("USER TERM SHELL COLUMNS LINES C_CTYPE XTERM_LOCALE DISPLAY "
             "SSH_CLIENT SSH_CONNECTION SSH_TTY HOME HOSTNAME PWD MAIL LANG "
             "PWD UID USER_ID EDITOR LOGNAME".split())))
        rstr += bytes(''.join((chr(3), IAC, SE)))
        self.ENV_REQUESTED = True
        self.send_str(rstr)

    def request_do_ttype(self):
        """
        Begins TERMINAL-TYPE negotiation
        """
        if self.check_remote_option(TTYPE) in (False, UNKNOWN):
            self._iac_do(TTYPE)
            self._note_reply_pending(TTYPE, True)

    def request_ttype(self):
        """
        Sends IAC SB TTYPE SEND IAC SE
        """
        self.send_str(bytes(''.join((
            IAC, SB, TTYPE, SEND, IAC, SE))))

    def send_ready(self):
        """
        Return True if any data is buffered for sending (screen output).
        """
        return bool(0 != self.send_buffer.__len__())

    def input_ready(self):
        """
        Return True if any data is buffered for reading (keyboard input).
        """
        return bool(0 != self.recv_buffer.__len__())

    def socket_send(self):
        """
        Called by TelnetServer.poll() when send data is ready.  Send any
        data buffered, trim self.send_buffer to bytes sent, and return number
        of bytes sent. throws Disconnected
        """
        if not self.send_ready():
            warnings.warn('socket_send() called on empty buffer',
                          RuntimeWarning, 2)
            return 0
        ready_bytes = bytes(''.join(self.send_buffer))
        self.send_buffer = array.array('c')

        def send(send_bytes):
            """
            throws x84.bbs.exception.Disconnected on sock.send err
            """
            try:
                return self.sock.send(send_bytes)
            except socket.error as err:
                if err[0] == 11:
                    warnings.warn('%s: %s (bandwidth exceed)' % (
                                self.addrport(), err[1],), RuntimeWarning, 2)
                else:
                    raise Disconnected(
                        'socket send %d: %s' % (err[0], err[1],))

        sent = send(ready_bytes)
        if sent < len(ready_bytes):
            # re-buffer data that could not be pushed to socket;
            self.send_buffer.fromstring(ready_bytes[sent:])
        else:
            # When a process has completed sending data to an NVT printer
            # and has no queued input from the NVT keyboard for further
            # processing (i.e., when a process at one end of a TELNET
            # connection cannot proceed without input from the other end),
            # the process must transmit the TELNET Go Ahead (GA) command.
            if (not self.input_ready()
                    and self.check_local_option(SGA) is False
                    and not self._check_reply_pending(SGA)):
                sent += send(bytes(''.join((IAC, GA))))
        return sent

    def socket_recv(self):
        """
        Called by TelnetServer.poll() when recv data is ready.  Read any
        data on socket, processing telnet commands, and buffering all
        other bytestrings to self.recv_buffer.  If data is not received,
        or the connection is closed, x84.bbs.exception.Disconnected is
        raised.
        """
        recv = 0
        try:
            data = self.sock.recv(self.BLOCKSIZE_RECV)
            recv = len(data)
            if 0 == recv:
                raise Disconnected('Closed by client')
        except socket.error as err:
            raise Disconnected('socket errno %d: %s' % (err[0], err[1],))
        self.bytes_received += recv
        self.last_input_time = time.time()

        ## Test for telnet commands, non-telnet bytes
        ## are pushed to self.recv_buffer (side-effect),
        for byte in data:
            self._iac_sniffer(byte)
        return recv

    def _recv_byte(self, byte):
        """
        Buffer non-telnet commands bytestrings into recv_buffer.
        """
        self.recv_buffer.fromstring(byte)

    def _iac_sniffer(self, byte):
        """
        Watches incomming data for Telnet IAC sequences.
        Passes the data, if any, with the IAC commands stripped to
        _recv_byte().
        """
        ## Are we not currently in an IAC sequence coming from the DE?
        if self.telnet_got_iac is False:
            if byte == IAC:
                self.telnet_got_iac = True
            ## Are we currenty in a sub-negotion?
            elif self.telnet_got_sb is True:
                self.telnet_sb_buffer.fromstring(byte)
                ## Sanity check on length
                if len(self.telnet_sb_buffer) >= self.SB_MAXLEN:
                    raise Disconnected('sub-negotiation buffer filled')
            else:
                ## Just a normal NVT character
                self._recv_byte(byte)
            return

        ## Did we get sent a second IAC?
        if byte == IAC and self.telnet_got_sb is True:
            ## Must be an escaped 255 (IAC + IAC)
            self.telnet_sb_buffer.fromstring(byte)
            self.telnet_got_iac = False
        ## Do we already have an IAC + CMD?
        elif self.telnet_got_cmd is not None:
            ## Yes, so handle the option
            self._three_byte_cmd(byte)
        ## We have IAC but no CMD
        else:
            ## Is this the middle byte of a three-byte command?
            if byte in (DO, DONT, WILL, WONT):
                self.telnet_got_cmd = byte
            else:
                ## Nope, must be a two-byte command
                self._two_byte_cmd(byte)

    def _two_byte_cmd(self, cmd):
        """
        Handle incoming Telnet commands that are two bytes long.
        """
        # logger.debug ('recv _two_byte_cmd %s', name_option(cmd),)
        if cmd == SB:
            ## Begin capturing a sub-negotiation string
            self.telnet_got_sb = True
            self.telnet_sb_buffer = array.array('c')
        elif cmd == SE:
            ## Stop capturing a sub-negotiation string
            self.telnet_got_sb = False
            self._sb_decoder()
        elif cmd == IP:
            self.deactivate()
            logger.debug('%s Interrupt; closing.', self.addrport())
        elif cmd == AO:
            flushed = len(self.recv_buffer)
            self.recv_buffer = array.array('c')
            logger.debug('Abort Output (AO); %s bytes discarded.', flushed)
        elif cmd == AYT:
            self.send_str(bytes('\b'))
            logger.debug('Are You There (AYT); "\\b" sent.')
        elif cmd == EC:
            self.recv_buffer.fromstring('\b')
            logger.debug('Erase Character (EC); "\\b" queued.')
        elif cmd == EL:
            logger.warn('Erase Line (EC) received; ignored.')
        elif cmd == GA:
            logger.warn('Go Ahead (GA) received; ignored.')
        elif cmd == NOP:
            logger.debug('NUL ignored.')
        elif cmd == DM:
            logger.warn('Data Mark (DM) received; ignored.')
        elif cmd == BRK:
            logger.warn('Break (BRK) received; ignored.')
        else:
            logger.error('_two_byte_cmd invalid: %r', cmd)
        self.telnet_got_iac = False
        self.telnet_got_cmd = None

    def _three_byte_cmd(self, option):
        """
        Handle incoming Telnet commmands that are three bytes long.
        """
        cmd = self.telnet_got_cmd
        logger.debug('recv IAC %s %s', name_option(cmd), name_option(option))
        # Incoming DO's and DONT's refer to the status of this end

        if cmd == DO:
            self._handle_do(option)
        elif cmd == DONT:
            self._handle_dont(option)
        elif cmd == WILL:
            self._handle_will(option)
        elif cmd == WONT:
            self._handle_wont(option)
        else:
            logger.warn('%s: unhandled _three_byte_cmd: %s.',
                        self.addrport(), name_option(option))
        self.telnet_got_iac = False
        self.telnet_got_cmd = None

    def _handle_do(self, option):
        """
        Process a DO command option received by DE.
        """
        # pylint: disable=R0912
        #         TelnetClient._handle_do: Too many branches (13/12)
        # if any pending WILL options were send, they have been received
        self._note_reply_pending(option, False)
        if option == ECHO:
            # DE requests us to echo their input
            if self.check_local_option(ECHO) is not True:
                self._note_local_option(ECHO, True)
                self._iac_will(ECHO)
        elif option == BINARY:
            # DE requests to recv BINARY
            if self.check_local_option(BINARY) is not True:
                self._note_local_option(BINARY, True)
                self._iac_will(BINARY)
        elif option == SGA:
            # DE wants us to supress go-ahead
            if self.check_local_option(SGA) is not True:
                self._note_local_option(SGA, True)
                # always send DO SGA after WILL SGA, requesting the DE
                # also supress their go-ahead. this order seems to be the
                # 'magic sequence' to disable linemode on certain clients
                self._iac_will(SGA)
                self._iac_do(SGA)
        elif option == LINEMODE:
            # DE wants to do linemode editing
            # denied
            if self.check_local_option(option) is not False:
                self._note_local_option(option, False)
                self._iac_wont(LINEMODE)
        elif option == ENCRYPT:
            # DE is willing to receive encrypted data
            # denied
            if self.check_local_option(option) is not False:
                self._note_local_option(option, False)
                # let DE know we refuse to send encrypted data.
                self._iac_wont(ENCRYPT)
        elif option == STATUS:
            # DE wants to know if we support STATUS,
            if self.check_local_option(option) is not True:
                self._note_local_option(option, True)
                self._iac_will(STATUS)
                self._send_status()
        else:
            if self.check_local_option(option) is UNKNOWN:
                self._note_local_option(option, False)
                logger.warn('%s: unhandled do: %s.',
                            self.addrport(), name_option(option))
                self._iac_wont(option)

    def _send_status(self):
        """
        Process a DO STATUS sub-negotiation received by DE. (rfc859)
        """
        # warning:
        rstr = bytes(''.join((IAC, SB, STATUS, IS)))
        for opt, status in self.telnet_opt_dict.items():
            # my_want_state_is_will
            if status.local_option is True:
                logger.debug('send WILL %s', name_option(opt))
                rstr += bytes(''.join((WILL, opt)))
            elif status.reply_pending is True and opt in (ECHO, SGA):
                logger.debug('send WILL %s (want)', name_option(opt))
                rstr += bytes(''.join((WILL, opt)))
            # his_want_state_is_will
            elif status.remote_option is True:
                logger.debug('send DO %s', name_option(opt))
                rstr += bytes(''.join((DO, opt)))
            elif (status.reply_pending is True
                    and opt in (NEW_ENVIRON, NAWS, TTYPE)):
                logger.debug('send DO %s (want)', name_option(opt))
                rstr += bytes(''.join((DO, opt)))
        rstr += bytes(''.join((IAC, SE)))
        logger.debug('send %s', ' '.join(name_option(opt) for opt in rstr))
        self.send_str(rstr)

    def _handle_dont(self, option):
        """
        Process a DONT command option received by DE.
        """
        self._note_reply_pending(option, False)
        self._note_reply_pending(option, False)
        if option == ECHO:
            # client demands we do not echo
            if self.check_local_option(ECHO) is not False:
                self._note_local_option(ECHO, False)
                self._iac_wont(ECHO)  # agree
        elif option == BINARY:
            # client demands no binary mode
            if self.check_local_option(BINARY) is not False:
                self._note_local_option(BINARY, False)
                self._iac_wont(BINARY)
        elif option == SGA:
            # DE demands that we start or continue transmitting
            # GAs (go-aheads) when transmitting data.
            if self.check_local_option(SGA) is not False:
                self._note_local_option(SGA, False)
                self._iac_wont(SGA)
        elif option == LINEMODE:
            # client demands no linemode.
            if self.check_remote_option(LINEMODE) is not False:
                self._note_remote_option(LINEMODE, False)
                self._iac_wont(LINEMODE)
        else:
            logger.warn('%s: unhandled dont: %s.',
                        self.addrport(), name_option(option))

    def _handle_will(self, option):
        """
        Process a WILL command option received by DE.
        """
        # pylint: disable=R0912
        #        Too many branches (19/12)
        self._note_reply_pending(option, False)
        if option == ECHO:
            raise Disconnected(
                'Refuse WILL ECHO by client, closing connection.')
        elif option == BINARY:
            if self.check_remote_option(BINARY) is not True:
                self._note_remote_option(BINARY, True)
                # agree to use BINARY
                self._iac_do(BINARY)
        elif option == NAWS:
            if self.check_remote_option(NAWS) is not True:
                self._note_remote_option(NAWS, True)
                self._note_local_option(NAWS, True)
                # agree to use NAWS, / go ahead ?
                self._iac_do(NAWS)
        elif option == STATUS:
            if self.check_remote_option(STATUS) is not True:
                self._note_remote_option(STATUS, True)
                self.send_str(bytes(''.join((
                    IAC, SB, STATUS, SEND, IAC, SE))))  # go ahead
        elif option in UNSUPPORTED_WILL:
            if self.check_remote_option(option) is not False:
                # let DE know we refuse to do linemode, encryption, etc.
                self._iac_dont(option)
        elif option == SGA:
            #  IAC WILL SUPPRESS-GO-AHEAD
            #
            # The sender of this command requests permission to begin
            # suppressing transmission of the TELNET GO AHEAD (GA)
            # character when transmitting data characters, or the
            # sender of this command confirms it will now begin suppressing
            # transmission of GAs with transmitted data characters.
            if self.check_remote_option(SGA) is not True:
                # sender of this command confirms that the sender of data
                # is expected to suppress transmission of GAs.
                self._iac_do(SGA)
                self._note_remote_option(SGA, True)
        elif option == NEW_ENVIRON:
            if self.check_remote_option(NEW_ENVIRON) in (False, UNKNOWN):
                self._note_remote_option(NEW_ENVIRON, True)
                self.request_env()
            self._note_local_option(NEW_ENVIRON, True)
        elif option == XDISPLOC:
            # if they want to send it, go ahead.
            if self.check_remote_option(XDISPLOC):
                self._note_remote_option(XDISPLOC, True)
                self._iac_do(XDISPLOC)
                self.send_str(bytes(''.join((
                    IAC, SB, XDISPLOC, SEND, IAC, SE))))
        elif option == TTYPE:
            if self.check_remote_option(TTYPE) in (False, UNKNOWN):
                self._note_remote_option(TTYPE, True)
                self.request_ttype()
        else:
            logger.warn('%s: unhandled will: %r (ignored).',
                        self.addrport(), name_option(option))

    def _handle_wont(self, option):
        """
        Process a WONT command option received by DE.
        """
        # pylint: disable=R0912
        #         TelnetClient._handle_wont: Too many branches (13/12)
        self._note_reply_pending(option, False)
        if option == ECHO:
            if self.check_remote_option(ECHO) in (True, UNKNOWN):
                self._note_remote_option(ECHO, False)
                self._iac_dont(ECHO)
        elif option == BINARY:
            # client demands no binary mode
            if self.check_remote_option(BINARY) in (True, UNKNOWN):
                self._note_remote_option(BINARY, False)
                self._iac_dont(BINARY)
        elif option == SGA:
            if self._check_reply_pending(SGA):
                self._note_reply_pending(SGA, False)
                self._note_remote_option(SGA, False)
            elif self.check_remote_option(SGA) in (True, UNKNOWN):
                self._note_remote_option(SGA, False)
                self._iac_dont(SGA)
        elif option == TTYPE:
            if self._check_reply_pending(TTYPE):
                self._note_reply_pending(TTYPE, False)
                self._note_remote_option(TTYPE, False)
            elif self.check_remote_option(TTYPE) in (True, UNKNOWN):
                self._note_remote_option(TTYPE, False)
                self._iac_dont(TTYPE)
        elif option in (NEW_ENVIRON, NAWS):
            if self._check_reply_pending(option):
                self._note_reply_pending(option, False)
                self._note_remote_option(option, False)
            elif self.check_remote_option(option) in (True, UNKNOWN):
                self._note_remote_option(option, False)
        else:
            logger.debug('%s: unhandled wont: %s.',
                         self.addrport(), name_option(option))
            self._note_remote_option(option, False)

    def _sb_decoder(self):
        """
        Figures out what to do with a received sub-negotiation block.
        """

        buf = self.telnet_sb_buffer
        if 0 == len(buf):
            logger.error('nil SB')
            return
        logger.debug('recv SB: %s %s',
                     name_option(buf[0]),
                     'IS %r' % (buf[2:],) if len(buf) > 1 and buf[1] is IS
                     else repr(buf[1:]))
        if 1 == len(buf) and buf[0] == chr(0):
            logger.error('0nil SB')
            return
        elif (TTYPE, IS) == (buf[0], buf[1]):
            self._sb_ttype(buf[2:].tostring())
        elif (XDISPLOC, IS) == (buf[0], buf[1]):
            self._sb_xdisploc(buf[2:].tostring())
        elif (NEW_ENVIRON, IS) == (buf[0], buf[1]):
            self._sb_env(buf[2:].tostring())
        elif NAWS == buf[0]:
            self._sb_naws(buf)
        elif (STATUS, SEND) == (buf[0], buf[1]):
            self._send_status()
        else:
            logger.error('unsupported subnegotiation, %s: %r',
                         name_option(buf[0]), buf,)
        self.telnet_sb_buffer = ''

    def _sb_xdisploc(self, bytestring):
        """
        Process incoming subnegotiation XDISPLOC
        """
        prev_display = self.env.get('DISPLAY', None)
        if prev_display is None:
            logger.info("env['DISPLAY'] = %r.", bytestring)
        elif prev_display != bytestring:
            logger.info("env['DISPLAY'] = %r by XDISPLOC was:%s.",
                        bytestring, prev_display)
        else:
            logger.debug('XDSIPLOC ignored (DISPLAY already set).')
        self.env['DISPLAY'] = bytestring

    def _sb_ttype(self, bytestring):
        """
        Processes incoming subnegotiation TTYPE
        """
        term_str = bytestring.lower().strip()
        while term_str.endswith('\x00'):
            term_str = term_str[:-1]  # netrunner did this ..
        prev_term = self.env.get('TERM', None)
        if prev_term is None:
            logger.info("env['TERM'] = %r.", term_str)
        elif prev_term != term_str:
            logger.info("env['TERM'] = %r by TTYPE%s.", term_str,
                        ', was: %s' % (prev_term,)
                        if prev_term != 'unknown' else '')
        else:
            logger.debug('TTYPE ignored (TERM already set).')
        self.env['TERM'] = term_str

    def _sb_env(self, bytestring):
        """
        Processes incoming subnegotiation NEW_ENVIRON
        """
        breaks = list([idx for (idx, byte) in enumerate(bytestring)
                       if byte in (chr(0), chr(3))])
        for start, end in zip(breaks, breaks[1:]):
            pair = bytestring[start + 1:end].split(chr(1))
            if len(pair) == 1:
                if (pair[0] in self.env
                        and pair[0] not in ('LINES', 'COLUMNS', 'TERM')):
                    logger.warn("del env[%r]", pair[0])
                    del self.env[pair[0]]
            elif len(pair) == 2:
                if pair[0] == 'TERM':
                    pair[1] = pair[1].lower()
                overwrite = (pair[0] == 'TERM'
                             and self.env['TERM'] == 'unknown')
                if (not pair[0] in self.env or overwrite):
                    logger.info('env[%r] = %r', pair[0], pair[1])
                    self.env[pair[0]] = pair[1]
                elif pair[1] == self.env[pair[0]]:
                    logger.debug('env[%r] repeated', pair[0])
                else:
                    logger.warn('%s=%s; conflicting value %s ignored.',
                                pair[0], self.env[pair[0]], pair[1])
            else:
                logger.error('client NEW_ENVIRON; invalid %r', pair)
        self.ENV_REPLIED = True

    def _sb_naws(self, charbuf):
        """
        Processes incoming subnegotiation NAWS
        """
        if 5 != len(charbuf):
            logger.error('%s: bad length in NAWS buf (%d)',
                         self.addrport(), len(charbuf),)
            return

        columns = str((256 * ord(charbuf[1])) + ord(charbuf[2]))
        rows = str((256 * ord(charbuf[3])) + ord(charbuf[4]))
        if (self.env.get('LINES', None) == rows
                and self.env.get('COLUMNS', None) == columns):
            logger.debug('%s: NAWS repeated', self.addrport())
        else:
            self.env['LINES'] = str(rows)
            self.env['COLUMNS'] = str(columns)
            logger.debug('%s: NAWS is %sx%s',
                         self.addrport(), columns, rows)
            if self.on_naws is not None:
                self.on_naws(self)

    #---[ State Juggling for Telnet Options ]----------------------------------
    ## Sometimes verbiage is tricky.  I use 'note' rather than 'set' here
    ## because (to me) set infers something happened.
    #@debug_option
    def check_local_option(self, option):
        """
        Test the status of local negotiated Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].local_option

    #@debug_option
    def _note_local_option(self, option, state):
        """
        Record the status of local negotiated Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].local_option = state

    #@debug_option
    def check_remote_option(self, option):
        """
        Test the status of remote negotiated Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].remote_option

    #@debug_option
    def _note_remote_option(self, option, state):
        """
        Record the status of local negotiated Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].remote_option = state

    #@debug_option
    def _check_reply_pending(self, option):
        """
        Test the status of requested Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        return self.telnet_opt_dict[option].reply_pending

    #@debug_option
    def _note_reply_pending(self, option, state):
        """
        Record the status of requested Telnet options.
        """
        if not option in self.telnet_opt_dict:
            self.telnet_opt_dict[option] = TelnetOption()
        self.telnet_opt_dict[option].reply_pending = state

    #---[ Telnet Command Shortcuts ]-------------------------------------------
    #@debug_option
    def _iac_do(self, option):
        """
        Send a Telnet IAC "DO" sequence.
        """
        logger.debug('send IAC DO %s', name_option(option))
        self.send_str(bytes(''.join((IAC, DO, option))))

    #@debug_option
    def _iac_dont(self, option):
        """
        Send a Telnet IAC "DONT" sequence.
        """
        logger.debug('send IAC DONT %s', name_option(option))
        self.send_str(bytes(''.join((IAC, DONT, option))))

    #@debug_option
    def _iac_will(self, option):
        """
        Send a Telnet IAC "WILL" sequence.
        """
        logger.debug('send IAC WILL %s', name_option(option))
        self.send_str(bytes(''.join((IAC, WILL, option))))

    #@debug_option
    def _iac_wont(self, option):
        """
        Send a Telnet IAC "WONT" sequence.
        """
        logger.debug('send IAC WONT %s', name_option(option))
        self.send_str(bytes(''.join((IAC, WONT, option))))



""" Who's online script for X/84, https://github.com/jquast/x84 """
import time
SELF_ID = -1
POLL_KEY = 0.25  # blocking ;; how often to poll keyboard
POLL_INF = 2.00  # seconds elapsed until re-ask clients for more details
POLL_AYT = 4.00  # seconds elapsed until global 'are you there?' is checked,
POLL_OUT = 0.50  # seconds elapsed before screen updates


def request_info(sid):
    """ Send info-req event to target session id ``sid``. """
    from x84.bbs import getsession
    session = getsession()
    session.send_event('route', (sid, 'info-req', session.sid,))


def banner():
    """ Returns string suitable for displaying banner """
    from x84.bbs import getterminal
    term = getterminal()
    return u''.join((
        u''.center((term.width / 2) - 3),
        term.green_underline('.'),
        term.green_bold_underline('.'),
        term.underline('.'),
        term.bold_green(" whO'S ONliNE"),))


def describe(sessions):
    """
    Returns unicode string suitable for describing the activity of
    session id's of array ``sessions``.
    """
    from x84.bbs import getsession, getterminal, ini
    slen = lambda sessions: len(u'%d' % (len(sessions),))
    session, term = getsession(), getterminal()
    max_user = ini.CFG.getint('nua', 'max_user')
    return u'\r\n'.join(([u''.join((
        u'%*d' % (4 + slen(sessions), node),
        u'%4is' % (attrs.get('idle', 0),), u' ',
        (term.bold_green(u'%-*s' % (max_user, (
        u'** diSCONNECtEd' if 'delete' in attrs
        else attrs.get('handle', u'** CONNECtiNG')),)
        ) if attrs.get('handle', u'') != session.user.handle
            else term.green(u'%-*s' % (max_user, session.user.handle))),
        term.green(u' - '),
        term.bold_green((attrs.get('activity', u''))
                        if attrs.get('sid') != session.sid else
                        term.bold_black(session.activity)),
    )) for node, (_sid, attrs) in get_nodes(sessions)]))


def get_nodes(sessions):
    """ Given an array of sessions, assign an arbitrary 'node' number """
    return enumerate(sorted(sessions.items()))


def heading(sessions):
    """
    Given an array of sessions, return string suitable for display heading.
    """
    from x84.bbs import getterminal, ini
    slen = lambda sessions: len(u'%d' % (len(sessions),))
    term = getterminal()
    max_user = ini.CFG.getint('nua', 'max_user')
    return u'\r\n'.join((
        u'\r\n'.join([pline.center(term.width)
                      for pline in prompt().splitlines()]),
        u'\r\n',
        term.green_underline(u''.join((
            'node'.rjust(4 + slen(sessions)),
            'idle'.rjust(5),
            ' handle'.ljust(max_user + 3),
            'activity',))),))


def prompt():
    """
    Return string suitable for displaying prompt and available commands.
    """
    from x84.bbs import getsession, getterminal, Ansi
    session, term = getsession(), getterminal()
    decorate = lambda key, desc: u''.join((
        u'(', term.green_underline(key,),
        u')', term.reverse_green(desc.split()[0]), u' ',
        u' '.join(desc.split()[1:]), u' ',))
    return Ansi(u''.join((
        u' ' * 2,
        term.green_reverse(':keys'), u' ',
        decorate('c', 'hAt USR'),
        decorate('s', 'ENd MSG'),
        (u''.join((
        decorate('p', 'lAYbACk REC'),
        decorate('w', 'AtCh liVE'),
        decorate('d', 'iSCONNECt SiD'),
        decorate('e', 'diT USR'),
        decorate('v', 'iEW SiD AttRS'),
        u' ',)) if 'sysop' in session.user.groups else u''),
        decorate('Escape/q', 'Uit'),
        decorate('Spacebar', 'REfRESh'),
    ))).wrap(int(term.width * .7), indent=u' ' * 8)


def get_node(sessions):
    """ Prompt user for session node, Returns node & session attributes. """
    from x84.bbs import ini, LineEditor, echo
    max_user = ini.CFG.getint('nua', 'max_user')
    invalid = u'\r\ninvalid.'
    echo(u'\r\n\r\nNOdE: ')
    node = LineEditor(max_user).read()
    if node is None or 0 == len(node):
        return (None, None)
    try:
        node = int(node)
    except ValueError:
        echo(invalid)
        return (None, None)
    for tgt_node, (_sid, attrs) in get_nodes(sessions):
        if tgt_node == node:
            return (tgt_node, attrs)
    echo(invalid)
    return (None, None)


def edit(sessions):
    """ Prompt for node and gosub profile.py script for user of target session.
    """
    from x84.bbs import gosub
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        gosub('profile', tgt_session['handle'])
        return True


def playback(sessions):
    """ Prompt for node and gosub ttyplay script for ttyrec of target session.
    """
    from x84.bbs import gosub
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        gosub('ttyplay', tgt_session['ttyrec'])
        return True


def watch(sessions):
    """
    Prompt for node and gosub ttyplay script for ttyrec of target session,
    with 'peek' boolean set to True.
    """
    from x84.bbs import gosub
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        gosub('ttyplay', tgt_session['ttyrec'], True)
        return True


def chat(sessions):
    """
    Prompt for node and page target session for chat.
    Sysop will send session id of -1, indicating the chat is forced.
    """
    from x84.bbs import gosub, getsession
    session = getsession()
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        # page other user,
        channel = tgt_session['sid']
        sender = (session.user.handle
                  if not 'sysop' in session.user.groups else -1)
        session.send_event('route', (
            tgt_session['sid'], 'page', channel, sender))
        gosub('chat', channel)
        return True


def view(sessions):
    """
    Prompt for node and view session details of target session.
    """
    from x84.bbs import echo, getterminal
    term = getterminal()
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        maxlen = max([len(key) for key in tgt_session.keys()])
        echo(u''.join((
            u'\r\n\r\n',
            u'\r\n'.join(['%s%s %s' % (
                term.bold('%*s' % (maxlen, key)),
                term.bold_green(':'),
                term.green(str(value)),)
                for key, value in sorted(tgt_session.items())]),
        )))
        return True


def disconnect(sessions):
    """
    Prompt for node and disconnect target session.
    """
    from x84.bbs import getsession
    session = getsession()
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        session.send_event('remote-disconnect', tgt_session['sid'])
        return True


def sendmsg(sessions):
    """
    Prompt for node and gosub 'writemsg' with recipient set to target user.
    """
    from x84.bbs import gosub, Msg
    (node, tgt_session) = get_node(sessions)
    if node is not None:
        msg = Msg()
        msg.recipient = tgt_session['handle']
        msg.tags.add('private')
        gosub('writemsg', msg)
        return True


def main():
    """ Main procedure. """
    # pylint: disable=R0912,R0914,R0915
    #         Too many branches
    #         Too many local variables
    #         Too many statements
    from x84.bbs import getsession, getterminal, getch, echo
    session, term = getsession(), getterminal()
    ayt_lastfresh = 0

    def broadcast_ayt(last_update):
        """ Globally boradcast 'are-you-there' request. """
        if time.time() - last_update > POLL_AYT:
            session.send_event('global', ('AYT', session.sid,))
            last_update = time.time()
        return last_update

    sessions = dict()
    dirty = time.time()
    cur_row = 0
    while True:
        ayt_lastfresh = broadcast_ayt(ayt_lastfresh)
        inp = getch(POLL_KEY)
        if session.poll_event('refresh') or (
                inp in (u' ', term.KEY_REFRESH, unichr(12))):
            dirty = time.time()
            cur_row = 0
        elif inp in (u'q', 'Q', term.KEY_EXIT, unichr(27)):
            return
        elif inp in (u'c', 'C'):
            cur_row = 0 if chat(sessions) else cur_row
            dirty = time.time()
        elif inp in (u's', 'S'):
            cur_row = 0 if sendmsg(sessions) else cur_row
            dirty = time.time()
        elif inp is not None and 'sysop' in session.user.groups:
            if inp in (u'e', u'E'):
                cur_row = 0 if edit(sessions) else cur_row
                dirty = time.time()
            elif inp in (u'p', u'P'):
                cur_row = 0 if playback(sessions) else cur_row
                dirty = time.time()
            elif inp in (u'w', u'W'):
                cur_row = 0 if watch(sessions) else cur_row
                dirty = time.time()
            elif inp in (u'v', u'V'):
                cur_row = 0 if view(sessions) else cur_row + 3
                dirty = time.time()
            elif inp in (u'd', u'D'):
                disconnect(sessions)
                dirty = time.time()

        # add sessions that respond to AYT
        data = session.poll_event('ACK')
        if data is not None:
            sid, handle = data
            if sid in sessions:
                sessions[sid]['handle'] = handle
            else:
                sessions[sid] = dict((
                    ('handle', handle),
                    ('lastfresh', time.time()),
                    ('lastasked', time.time()),))
                dirty = time.time()
                echo(u'\a')

        # update sessions that respond to info-req
        data = session.poll_event('info-ack')
        if data is not None:
            sid, attrs = data
            if sessions.get(sid, dict()).get('activity') != attrs['activity']:
                # and refresh screen if activity changes
                dirty = time.time()
            sessions[sid] = attrs
            sessions[sid]['lastfresh'] = time.time()

        # update our own session
        sessions[SELF_ID] = session.info()
        sessions[SELF_ID]['lastfresh'] = time.time()

        # request that all sessions update if more stale than POLL_INF,
        # or is missing session info (only AYT replied so far!),
        # or has been displayed as 'Disconnected' (marked for deletion)
        for sid, attrs in sessions.items():
            if sid == SELF_ID:
                continue
            if attrs.get('idle', -1) == -1 or (
                    time.time() - attrs.get('lastfresh', 0) > POLL_INF
                    and time.time() - attrs.get('lastasked', 0) > POLL_INF):
                request_info(sid)
                attrs['lastasked'] = time.time()

        # prune users who haven't responded to AYT
        for sid, attrs in sessions.items():
            if time.time() - attrs['lastfresh'] > (POLL_AYT * 2):
                sessions[sid]['delete'] = 1
                dirty = time.time()

        if dirty is not None and time.time() - dirty > POLL_OUT:
            session.activity = u"Who's Online"
            otxt = describe(sessions)
            olen = len(otxt.splitlines())
            if 0 == cur_row or (cur_row + olen) >= term.height:
                otxt_b = banner()
                otxt_h = heading(sessions)
                cur_row = len(otxt_b.splitlines()) + len(otxt_h.splitlines())
                echo(u'\r\n'.join((u'\r\n\r\n', otxt_b, otxt_h, otxt)))
            else:
                echo(u''.join((
                    u'\r\n',
                    '-'.center(term.width).rstrip(),
                    u'\r\n')))
                echo(otxt)
            cur_row += olen
            dirty = None

        # delete disconnected sessions
        for sid, attrs in sessions.items()[:]:
            if attrs.get('delete', 0) == 1:
                del sessions[sid]



"""
Pager class for x/84, http://github.com/jquast/x84/
"""
from x84.bbs.ansiwin import AnsiWindow
from x84.bbs.output import Ansi
import logging

VI_KEYSET = {
    'refresh': [unichr(12), ],
    'home': [],
    'end': [],
    'up': [u'k', u'K'],
    'down': [u'j', u'J', u'\r'],
    'pgup': [u'b', u'B'],
    'pgdown': [u'f', u'F'],
    'exit': [u'q', u'Q', unichr(27), ],
}


class Pager(AnsiWindow):
    """
    Scrolling ansi viewert d
    """
    # pylint: disable=R0904,R0902
    #        Too many public methods (24/20)
    #        Too many instance attributes (11/7)

    def __init__(self, height, width, yloc, xloc):
        """
        Initialize a pager of height, width, y, and x position.
        """
        AnsiWindow.__init__(self, height, width, yloc, xloc)
        self._position = self._position_last = 0
        self._quit = False
        self.content = list()
        self.keyset = VI_KEYSET.copy()
        self.init_keystrokes()

    @property
    def quit(self):
        """
        Returns: True if a terminating or quit character was handled by
        process_keystroke(), such as the escape key, or 'q' by default.
        """
        return self._quit

    @property
    def position_last(self):
        """
        Previous position before last move
        """
        return self._position_last

    @property
    def position(self):
        """
        Index of content buffer displayed at top of window.
        """
        return self._position

    @position.setter
    def position(self, pos):
        # pylint: disable=C0111
        #         Missing docstring
        self._position_last = self.position
        self._position = pos
        # bounds check
        if self._position < 0:
            self._position = 0
        if self._position > self.bottom:
            self._position = self.bottom
        self.moved = (self._position_last != self._position)

    @property
    def visible_content(self):
        """
        Returns content that is visible in window
        """
        return self.content[self.position:self.position + self.visible_height]

    @property
    def visible_bottom(self):
        """
        Returns bottom-most window row that contains content
        """
        if self.bottom < self.visible_height:
            return self.bottom
        return len(self.visible_content) - 1

    @property
    def bottom(self):
        """
        Returns bottom-most position that contains content
        """
        return max(0, len(self.content) - self.visible_height)

    def init_keystrokes(self):
        """
        This initializer sets keys appropriate for navigation.
        override or inherit this method to create a common keyset.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        self.keyset['home'].append(term.KEY_HOME)
        self.keyset['end'].append(term.KEY_END)
        self.keyset['pgup'].append(term.KEY_PPAGE)
        self.keyset['pgdown'].append(term.KEY_NPAGE)
        self.keyset['up'].append(term.KEY_UP)
        self.keyset['down'].append(term.KEY_DOWN)
        self.keyset['down'].append(term.KEY_ENTER)
        self.keyset['exit'].append(term.KEY_EXIT)

    def process_keystroke(self, keystroke):
        """
        Process the keystroke received by run method and return terminal
        sequence suitable for refreshing when that keystroke modifies the
        window.
        """
        import x84.bbs.session
        self.moved = False
        rstr = u''
        if keystroke in self.keyset['refresh']:
            rstr += self.refresh()
        elif keystroke in self.keyset['up']:
            rstr += self.move_up()
        elif keystroke in self.keyset['down']:
            rstr += self.move_down()
        elif keystroke in self.keyset['home']:
            rstr += self.move_home()
        elif keystroke in self.keyset['end']:
            rstr += self.move_end()
        elif keystroke in self.keyset['pgup']:
            rstr += self.move_pgup()
        elif keystroke in self.keyset['pgdown']:
            rstr += self.move_pgdown()
        elif keystroke in self.keyset['exit']:
            self._quit = True
        else:
            logger = logging.getLogger()
            logger.debug(
                'unhandled, %r', keystroke if type(keystroke) is not int
                else x84.bbs.session.getterminal().keyname(keystroke))
        return rstr

    def read(self):
        """
        Reads input until ESCAPE key is pressed (Blocking).  Returns None.
        """
        from x84.bbs import getch
        from x84.bbs.session import getsession
        from x84.bbs.output import echo
        session = getsession()
        self._quit = False
        echo(self.refresh())
        while not self.quit:
            echo(self.process_keystroke(getch()))

    def move_home(self):
        """
        Scroll to top.
        """
        self.position = 0
        if self.moved:
            return self.refresh()
        return u''

    def move_end(self):
        """
        Scroll to bottom.
        """
        self.position = len(self.content) - self.visible_height
        if self.moved:
            return self.refresh()
        return u''

    def move_pgup(self, num=1):
        """
        Scroll up ``num`` pages.
        """
        self.position -= (num * (self.visible_height))
        return self.refresh() if self.moved else u''

    def move_pgdown(self, num=1):
        """
        Scroll down ``num`` pages.
        """
        self.position += (num * (self.visible_height))
        return self.refresh() if self.moved else u''

    def move_down(self, num=1):
        """
        Scroll down ``num`` rows.
        """
        self.position += num
        if self.moved:
            return self.refresh()
        return u''

    def move_up(self, num=1):
        """
        Scroll up ``num`` rows.
        """
        self.position -= num
        if self.moved:
            return self.refresh()
        return u''

    def refresh_row(self, row):
        """
        Return unicode string suitable for refreshing pager window at
        visible row.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        ucs = (Ansi(self.visible_content[row])
               if row < len(self.visible_content)
               else u'')
        return u''.join((
            term.normal,
            self.pos(row + self.ypadding, self.xpadding),
            self.align(ucs),
            term.normal))

    def refresh(self, start_row=0):
        """
        Return unicode string suitable for refreshing pager window from
        optional visible content row 'start_row' and downward. This can be
        useful if only the last line is modified; or in an 'insert' operation,
        only the last line need be refreshed.
        """
        import x84.bbs.session
        term = x84.bbs.session.getterminal()
        return u''.join(
            [term.normal] + [self.refresh_row(row)
                for row in range(start_row, len(self.visible_content))]
            + [term.normal])

    def update(self, ucs):
        """
        Update content buffer with '\n'-delimited lines of Ansi.
        """
        try:
            self.content = Ansi(Ansi(ucs).decode_pipe()).wrap(
                self.visible_width).splitlines()
        except AssertionError, err:
            # indeterminate length
            logger = logging.getLogger()
            logger.warn('%s in [%r]', err, ucs)
            self.content = ucs.split('\r\n')
        return self.refresh()

    def append(self, ucs):
        """
        Update content buffer with additional lines of ansi unicodes.
        """
        self.content.extend(Ansi(Ansi(ucs).decode_pipe())
                            .wrap(self.visible_width - 1).splitlines())
        return self.move_end() or self.refresh(self.bottom)



"""
Door package for x/84 BBS http://github.com/jquast/x84
"""
import resource
import termios
import logging
import select
import codecs
import struct
import time
import fcntl
import pty
import sys
import os
import re

class Dropfile(object):
    (DOORSYS, DOOR32, CALLINFOBBS, DORINFO) = range(4)
    DOORSYS_GM = 'GR' # graphics mode

    def __init__(self, filetype=None):
        assert filetype in (self.DOORSYS, self.DOOR32,
                self.CALLINFOBBS, self.DORINFO)
        self.filetype = filetype

    def save(self, folder):
        """ Save dropfile to folder """
        fp = codecs.open(os.path.join(folder, self.filename), 'w', 'ascii', 'replace')
        fp.write(self.__str__())
        fp.close()

    @property
    def node(self):
        from x84.bbs import getsession
        return getsession().node

    @property
    def location(self):
        from x84.bbs import getsession
        return getsession().user.location

    @property
    def fullname(self):
        from x84.bbs import getsession
        return '%s %s' % (
                getsession().user.handle,
                getsession().user.handle,)

    @property
    def securitylevel(self):
        from x84.bbs import getsession
        return 100 if getsession().user.is_sysop else 30

    @property
    def numcalls(self):
        from x84.bbs import getsession
        return getsession().user.calls

    @property
    def lastcall_date(self):
        from x84.bbs import getsession
        return time.strftime('%m/%d/%y',
                time.localtime(getsession().user.lastcall))

    @property
    def lastcall_time(self):
        from x84.bbs import getsession
        return time.strftime('%H:%M',
                time.localtime(getsession().user.lastcall))

    @property
    def time_used(self):
        from x84.bbs import getsession
        return int(time.time() - getsession().connect_time)

    @property
    def remaining_secs(self):
        return 256 * 60

    @property
    def remaining_mins(self):
        return 256

    @property
    def comport(self):
        return 'COM1'

    @property
    def comspeed(self):
        return 57600

    @property
    def comtype(self):
        return 0 ## Line 1 : Comm type (0=local, 1=serial, 2=telnet)

    @property
    def comhandle(self):
        return 0 ## Line 2 : Comm or socket handle

    @property
    def parity(self):
        return 8

    @property
    def password(self):
        return '<encrypted>'

    @property
    def pageheight(self):
        from x84.bbs import getterminal
        return getterminal().height

    @property
    def systemname(self):
        from x84.bbs import ini
        return ini.CFG.get('system', 'software')

    @property
    def xferprotocol(self):
        return 'X' # x-modem for now, we don't have any xfer code/prefs

    @property
    def usernum(self):
        from x84.bbs import getsession
        from x84.bbs.userbase import list_users
        try:
            return list_users().index(getsession().user.handle)
        except ValueError:
            return 999

    @property
    def sysopname(self):
        from x84.bbs import ini
        return ini.CFG.get('system', 'sysop')

    @property
    def alias(self):
        from x84.bbs import getsession
        return getsession().user.handle

    @property
    def filename(self):
        if self.filetype == self.DOORSYS:
            return 'DOOR.SYS'
        elif self.filetype == self.DOOR32:
            return 'DOOR32.SYS'
        elif self.filetype == self.CALLINFOBBS:
            return 'CALLINFO.BBS'
        elif self.filetype == self.DORINFO:
            # n in DORINFOn.DEF is 1-9,0,a-z
            if self.node == 10:
                nodeid = '0'
            elif self.node < 10:
                nodeid = str(self.node)
            else:
                nodeid = chr(ord('a') + (self.node - 11))
                assert ord(nodeid) <= ord('z')
            return 'DORINFO%s.DEF' % (nodeid,)
        else:
            raise ValueError('filetype is unknown')

    def __str__(self):
        if self.filetype == self.DOORSYS:
            return self.get_doorsys()
        elif self.filetype == self.DOOR32:
            return self.get_door32()
        elif self.filetype == self.CALLINFOBBS:
            return self.get_callinfo()
        elif self.filetype == self.DORINFO:
            return self.get_dorinfo()
        else:
            raise ValueError('filetype is unknown')

    def get_doorsys():
        return (u'%s:\r\n%d\r\n'  # comport, comspeed
                '%d\r\n%d\r\n'  # parity, node
                '%d\r\nY\r\n'  # comspeed, screen?
                'Y\r\nY\r\nY\r\n'  # printer? pager alarm? caller alarm?
                '%s\r\n%s\r\n'  # fullname, location
                '123-456-7890\r\n123-456-7890\r\n'  # phone numbers
                '%s\r\n%d\r\n%d\r\n'  # password, security level, numcalls
                '%d\r\n%d\r\n%d\r\n'  # lastcall, remaining (mins, secs)
                'NG\r\n%d\r\nN\r\n'  # graphics mode, page length, expert mode
                '1,2,3,4,5,6,7\r\n1\r\n'  # conferences, conf. sel, exp. date
                '01/01/99\r\n'  # exp. date
                '%s\r\n%s\r\n%s\r\n'  # user number, def. xfer protocol,
                '0\r\n0\r\n'  # total #u/l, total #d/l
                '0\r\n9999999\r\n'  # daily d/l limit return val/write val
                '01/01/2001\r\n'  # birthdate
                'C:\\XXX\r\nC:\\XXX\r\n'  # filepaths to bbs files ...
                '%s\r\n%s\r\n'  # sysop's name, user's alias
                '00:05\r\nY\r\n'  # "event time"?, error-correcting connection
                'Y\r\nY\r\n'  # is ANSI in NG mode? Use record locking?
                '7\r\n%d\r\n'  # default color .. time credits in minutes
                '09/09/99\r\n%s\r\n'  # last new file scan, time of call,
                '%s\r\n9999\r\n'  # time of last call, max daily files
                '0\r\n0\r\n0\r\n' # files, u/l Kb, d/l Kb today
                'None\r\n0\r\n0\n' # user comment, doors opened, msgs left
                % (
                    self.comport, self.comspeed, self.parity, self.node,
                    self.comspeed, self.fullname, self.location,
                    self.password, self.securitylevel, self.numcalls,
                    self.lastcall_date, self.remaining_secs,
                    self.remaining_mins, self.pageheight, self.usernum,
                    self.xferprotocol, self.sysopname, self.alias,
                    self.remaining_mins, self.lastcall_time,
                    self.lastcall_time))

    def get_door32():
        return (u'%d\r\n%d\r\n%d\r\n'  # comm type, handle, speed
                '%s\r\n%d\r\n%s\r\n'  # system name, user num, real name
                '%s\r\n%d\r\n%d\r\n'  # alias, security level, mins remain,
                '1\r\n%d\n'  # emulation ('ansi'), current node num,
                % (
                    self.comtype, self.comhandle, self.comspeed,
                    self.systemname, self.usernum, self.fullname,
                    self.alias, self.securitylevel, self.remaining_mins,
                    self.node))
    def get_callinfo():
        return (u'%s\r\n%d\r\n%s\r\n'  # user name, comspeed, location
                '%d\r\n%d\r\nCOLOR\r\n' # security level, mins remain, ansi?
                '%s\r\n%d\r\n%d\r\n'  # password, usernum, time_used,
                '01:23\r\n01:23 01/02/90\r\n'
                'ABCDEFGH\r\n0\r\n'
                '99\r\n0\r\n'
                '9999\r\n'  # 7 unknown fields,
                '123-456-7890\r\n01/01/90 02:34\r\n' # phone, unknown
                'NOVICE\r\n%s\r\n'  # expert mode, xfer protocol
                '01/01/90\r\n%d\r\n' # unknown, number of calls,
                '%d\r\n0\r\n' # lines per screen, ptr to new msgs?
                '0\r\n0\r\n' # total u/l, d/l
                '8  { Databits }\r\n' #  who knows
                'REMOTE\r\n%s\r\n' # LOCAL|REMOTE, comport,
                '%d\r\nFALSE\r\n' # comspeed, unknown,
                'Normal Connection\r\n' # unknown,
                '01/02/94 01:20\r\n0\r\n1\n' # unknown, "task #", "door #"
                % (self.alias, self.comspeed, self.location,
                    self.securitylevel, self.remaining_mins, self.password,
                    self.usernum, self.time_used, self.xferprotocol,
                    self.numcalls, self.pageheight, self.comspeed, ))

    def get_dorinfo(self):
        return (u'%s\r\n%s\r\n%s\r\n' # software, sysop fname, sysop lname,
                '%s\r\n%d\r\n0\r\n'  # com port, bps, "networked"?
                '%s\r\n%s\r\n%s\r\n' # user fname, user lname, user location
                '1\r\n%d\r\n%d\r\n' # term (ansi), security level, mins remain
                '-1\n'  # fossil (-1 = "using external serial driver"..)
                % (
                    self.systemname, self.sysopname, self.sysopname,
                    self.comport, self.comspeed,
                    self.alias, self.alias, self.location,
                    self.securitylevel, self.remaining_mins,))


class Door(object):
    """
    Spawns a subprocess and pipes input and output over bbs session.
    """
    # pylint: disable=R0903
    #        Too few public methods (1/2)
    time_ipoll = 0.05
    time_opoll = 0.05
    blocksize = 7680
    timeout = 1984
    master_fd = None

    def __init__(self, cmd='/bin/uname', args=(), env_lang='en_US.UTF-8',
                 env_term=None, env_path=None, env_home=None, cp437=False):
        # pylint: disable=R0913
        #        Too many arguments (7/5)
        """
        cmd, args = argv[0], argv[1:]
        lang, term, and env_path become LANG, TERM, and PATH environment
        variables. When env_term is None, the session terminal type is used.
        When env_path is None, the .ini 'env_path' value of section [door] is
        used.  When env_home is None, $HOME of the main process is used.
        """
        from x84.bbs import getsession, getterminal, ini
        self._session, self._term = getsession(), getterminal()
        # pylint: disable=R0913
        #        Too many arguments (7/5)
        self.cmd = cmd
        if type(args) is tuple:
            self.args = (self.cmd,) + args
        elif type(args) is list:
            self.args = [self.cmd,] + args
        else:
            raise ValueError, 'args must be tuple or list'
        self.env_lang = env_lang
        if env_term is None:
            self.env_term = self._session.env.get('TERM')
        else:
            self.env_term = env_term
        if env_path is None:
            self.env_path = ini.CFG.get('door', 'path')
        else:
            self.env_path = env_path
        if env_home is None:
            self.env_home = os.getenv('HOME')
        else:
            self.env_home = env_home
        self.env = None # add additional env variables ...
        self.cp437 = cp437
        self._utf8_decoder = codecs.getincrementaldecoder('utf8')()

    def run(self):
        """
        Begin door execution. pty.fork() is called, child process
        calls execvpe() while the parent process pipes telnet session
        IPC data to and from the slave pty until child process exits.
        """
        logger = logging.getLogger()
        env = dict() if self.env is None else self.env
        env.update({'LANG': self.env_lang,
               'TERM': self.env_term,
               'PATH': self.env_path,
               'HOME': self.env_home,
               'LINES': '%s' % (self._term.height,),
               'COLUMNS': '%s' % (self._term.width,)})
        logger.debug('os.execvpe(cmd=%r, args=%r, env=%r',
                self.cmd, self.args, env)
        try:
            pid, self.master_fd = pty.fork()
        except OSError, err:
            # too many open files, out of memory, no such file/directory
            logger.error('OSError in pty.fork(): %s', err)
            return

        # child process
        if pid == pty.CHILD:
            sys.stdout.flush()
            # send initial screen size
            fcntl.ioctl(sys.stdout.fileno(), termios.TIOCSWINSZ,
                        struct.pack('HHHH',
                            self._term.height, self._term.width, 0, 0))
            # we cannot log an exception, only print to stderr and have
            # it captured by the parent process; this is because our 'logger'
            # instance is dangerously forked, and any attempt to communicate
            # with multiprocessing pipes, loggers, etc. will cause the value
            # and state of many various file descriptors to become corrupted
            try:
                os.execvpe(self.cmd, self.args, env)
            except Exception as err:
                sys.stderr.write('%s\n' % (err,))
            os._exit(1)

        # parent process
        #
        # execute self._loop() and catch all i/o and o/s errors
        try:
            logger.info('exec/%s: %s', pid, ' '.join(self.args))
            self._loop()
        except IOError, err:
            logger.error('IOError: %s', err)
        except OSError, err:
            # match occurs on read() after child closed sys.stdout. (ok)
            if 'Errno 5' not in str(err):
                # otherwise log as an error,
                logger.error('OSError: %s', err)
        (pid, status) = os.waitpid(pid, 0)
        res = status >> 8
        if res != 0:
            logger.error('%s child %s exit %d', self.cmd, pid, res)
        else:
            logger.debug('%s exit', self.cmd)
        os.close(self.master_fd)
        return res

    def input_filter(self, data):
        """ When keyboard input is detected, this method may filter such input.
        """
        return data

    def output_filter(self, data):
        """ Given door output in bytes, if 'cp437' is specified in class
        constructor, convert to utf8 glyphs using cp437 encoding; otherwise
        decode output as utf8. """
        if self.cp437:
            return u''.join((CP437[ord(ch)] for ch in data))

        decoded = list()
        for num, byte in enumerate(data):
            final = ((num + 1) == len(data)
                     and not self._masterfd_isready())
            ucs = self._utf8_decoder.decode(byte, final)
            if ucs is not None:
                decoded.append(ucs)
        return u''.join(decoded)

    def _masterfd_isready(self):
        """
        returns True if bytes waiting on master fd, meaning
        this utf8 byte must really be the last for a while.
        """
        return self.master_fd != -1 and (self.master_fd in
                select.select([self.master_fd, ], (), (), 0)[0])

    def resize(self):
        logger = logging.getLogger()
        logger.debug('send TIOCSWINSZ: %dx%d',
                     self._term.width, self._term.height)
        fcntl.ioctl(self.master_fd, termios.TIOCSWINSZ,
                    struct.pack('HHHH',
                        self._term.height, self._term.width, 0, 0))

    def _loop(self):
        # pylint: disable=R0914
        #         Too many local variables (21/15)
        """
        Poll input and outpout of ptys,
        """
        from x84.bbs import echo
        from x84.bbs.cp437 import CP437
        logger = logging.getLogger()
        while True:
            # block up to self.time_opoll for screen output
            if self.master_fd == -1:
                # pty file descriptor closed by child, early termination!
                break
            rlist = (self.master_fd,)
            ret_tuple = select.select(rlist, (), (), self.time_opoll)
            if self.master_fd in ret_tuple[0]:
                data = os.read(self.master_fd, self.blocksize)
                if 0 == len(data):
                    break
                echo(self.output_filter(data))

            # block up to self.time_ipoll for keyboard input
            event, data = self._session.read_events(
                ('refresh', 'input',), self.time_ipoll)

            if event == 'refresh' and data[0] == 'resize':
                self.resize()

            elif event == 'input':
                data = self.input_filter(data)
                if 0 != len(data):
                    n_written = os.write(self.master_fd, data)
                    if n_written == 0:
                        logger.warn('fight 0-byte write; exit, right?')
                    if n_written != len(data):
                        # we wrote none or some of our keyboard input, but
                        # not all. re-buffer remaining bytes back into
                        # session for next poll
                        self._session.buffer_input(data[n_written:])
                        # XXX I've never actually seen this, though. It might
                        # require writing a sub-program that artificially
                        # hangs, such as time.sleep(99999) to assert correct
                        # behavior. Please report, should be ok ..
                        logger.warn('buffer_input(%r)!', data[n_written:])

class DOSDoor(Door):
    """ This Door-derived class removes the "report cursor position" query
    sequence, which is sent by DOSEMU on startup. It also removes the "switch
    to alternate screen mode" set and reset (blessings terminals provide this
    with the context manager, ala "with term.fullscreen():".

    It would appear that any early keyboard input received (esp. in response
    to "report cursor position") prior to DOOR execution in DOSEMU causes all
    input to be bitshifted and invalid and/or broken.

    This class should resolve such issues by ovveriding output_filter to
    remove such sequence, and input_filter which only allows input after a
    few seconds have passed.
    """
    RE_REPWITH_CLEAR = (
            r'\033\[('
                r'1;80H.*\033\[1;1H'
                r'|H\033\[2J'
                r'|\d+;1H.*\033\[1;1H'
                r')')
    RE_REPWITH_NONE = (
            r'\033\[('
            r'6n'
            r'|\?1049[lh]'
            r'|\d+;\d+r'
            r'|1;1H\033\[\dM)')
    START_BLOCK = 4.0

    def __init__(self, cmd='/bin/uname', args=(), env_lang='en_US.UTF-8',
                 env_term=None, env_path=None, env_home=None, cp437=False):
        Door.__init__(self, cmd, args,
                env_lang, env_term, env_path, env_home, cp437)
        self.check_winsize()
        self._stime = time.time()
        self._re_trim_clear = re.compile(self.RE_REPWITH_CLEAR, flags=re.DOTALL)
        self._re_trim_none = re.compile(self.RE_REPWITH_NONE, flags=re.DOTALL)
        self._replace_clear = (
                self._term.move(25, 0)
                + (u'\r\n' * 25)
                + self._term.home)

    def output_filter(self, data):
        data = Door.output_filter(self, data)
        if self._stime is not None and (
                time.time() - self._stime < self.START_BLOCK):
            data = re.sub(pattern=self._re_trim_clear,
                    repl=(self._replace_clear), string=data)
            data = re.sub(pattern=self._re_trim_none,
                    repl=u'\r\n', string=data)
        return data

    def input_filter(self, data):
        return data if time.time() - self._stime > self.START_BLOCK else ''

    def check_winsize(self):
        assert self._term.width >= 80, (
                'Terminal width must be greater than '
                '80 columns (IBM-PC dimensions). '
                'Please resize your window.')
        assert self._term.height >= 25, (
                'Terminal height must be greater than '
                '25 rows (IBM-PC dimensions). '
                'Please resize your window.')

    def resize(self):
        pass

    def run(self):
        """
        Begin door execution. pty.fork() is called, child process
        calls execvpe() while the parent process pipes telnet session
        IPC data to and from the slave pty until child process exits.

        On exit, DOSDoor flushes any keyboard input; DOSEMU appears to
        send various terminal reset sequences that may cause a reply to
        be received on input, and later as an invalid menu command.
        """
        from x84.bbs import echo
        echo('\r\n' * self._term.height)
        Door.run(self)

        # flush any previously decoded but unreceived keystrokes,
        # and any unprocessed input from telnet session not yet processed.
        self._term.kbflush()
        self._session.flush_event('input')

        # perform losless "cls" after dosemu exit; display is garbage
        echo(self._term.normal + u'\r\n' * self._term.height)



#! /usr/bin/env python
"""
Distribution file for x/84
"""
from distutils.core import setup
import os

setup(name='x84',
      version='1.0.8',
      description="Telnet server for UTF-8 and cp437 terminals.",
      long_description=open(os.path.join(os.path.dirname(__file__),
                                         'README.txt')).read(),
      author='Jeff Quast',
      author_email='contact@jeffquast.com',
      url='http://x84.rtfd.org/',
      keywords='telnet, terminal, server, ansi, bbs, mud, curses, utf8, cp437',
      license='ISC',
      packages=['x84', 'x84.default', 'x84.default.art', 'x84.bbs'],
      package_data={
          '': ['README.txt'],
          'x84.default': ['*.ans', '*.txt', ],
          'x84.default.art': ['*.asc', '*.ans', '*.txt'],
      },
      install_requires=[
          'requests >=1.1.0',
          'sauce >=0.1.1',
          'sqlitedict >=1.0.8'],
      scripts=['bin/x84'],
      classifiers=[
          'Environment :: Console',
          'Environment :: Console :: Curses',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: ISC License (ISCL)',
          'Natural Language :: English',
          'Operating System :: MacOS :: MacOS X',
          'Operating System :: POSIX',
          'Operating System :: POSIX :: BSD',
          'Operating System :: POSIX :: BSD :: FreeBSD',
          'Operating System :: POSIX :: BSD :: NetBSD',
          'Operating System :: POSIX :: BSD :: OpenBSD',
          'Operating System :: POSIX :: Linux',
          'Operating System :: POSIX :: SunOS/Solaris',
          'Operating System :: Unix',
          'Programming Language :: Python :: 2.6',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 2 :: Only',
          'Topic :: Artistic Software',
          'Topic :: Communications :: BBS',
          'Topic :: Software Development :: Libraries :: Application Frameworks',
          'Topic :: Software Development :: User Interfaces',
          'Topic :: System :: Console Fonts',
          'Topic :: Terminals',
          'Topic :: Terminals :: Telnet', ],
      )



"""
Session IPC for x/84, http://github.com/jquast/x84/
"""
import logging


class IPCLogHandler(logging.Handler):
    """
    Log handler that sends the log up the 'event pipe'. This is a rather novel
    solution that seems overlooked in documentation or exisiting code, try it!
    """
    def __init__(self, out_queue):
        """ Constructor method, requires multiprocessing.Pipe """
        logging.Handler.__init__(self)
        self.oqueue = out_queue
        self.session = None

    def emit(self, record):
        """
        emit log record via IPC output queue
        """
        try:
            e_inf = record.exc_info
            if e_inf:
                # side-effect: sets record.exc_text
                dummy = self.format(record)
                record.exc_info = None
                # pylint: disable=W0104
                #         Statement seems to have no effect
                dummy  # pflakes ;/
            record.handle = (self.session.handle
                             if self.session is not None else None)
            if self.session is not None:
                self.session.lock.acquire()
            self.oqueue.send(('logger', record))
            if self.session is not None:
                self.session.lock.release()
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            self.handleError(record)


class IPCStream(object):
    """
    Connect blessings 'stream' to 'child' output multiprocessing.Queue
    only write() is called by blessings.
    """
    # pylint: disable=R0903
    #         Too few public methods
    def __init__(self, out_queue, lock):
        self.oqueue = out_queue
        self.lock = lock
        self.is_a_tty = True

    def write(self, ucs, encoding='ascii'):
        """
        Sends unicode text to Pipe. Default encoding is 'ascii',
        which is unset only when used with blessings, which rarely
        writes directly to the stream.
        """
        self.lock.acquire()
        self.oqueue.send(('output', (ucs, encoding)))
        self.lock.release()



"""
Write public or private posts for x/84, https://github.com/jquast/x84/
"""


def banner():
    """ Display banner/art ... nothing for now """
    from x84.bbs import echo, getterminal
    term = getterminal()
    echo(u'\r\n\r\n')
    echo(term.bold_black(u'art needed ../'.center(term.width).rstrip()))
    echo(u'\r\n\r\n')


def display_msg(msg):
    """ Refresh screen, level indicates up to which step """
    from x84.bbs import getterminal, getsession, echo, Ansi
    session, term = getsession(), getterminal()
    body = msg.body.splitlines()
    echo(u'    AUthOR: ' + term.bold_yellow(msg.author) + u'\r\n\r\n')
    echo(u'   RECiPiENt: ')
    echo(term.yellow(msg.recipient
                     if msg.recipient is not None
                     else u'<(None)=All users>'))
    echo(u'\r\n\r\n')
    echo(u'     SUBjECt: ')
    echo(term.yellow(msg.subject))
    echo(u'\r\n\r\n')
    echo(u'        tAGS: ')
    echo(term.yellow(u', '.join(msg.tags)))
    echo(u'\r\n\r\n')
    echo(term.underline(u'        bOdY: '.ljust(term.width - 1)) + u'\r\n')
    echo(Ansi(u'\r\n'.join(body)).decode_pipe() + term.normal)
    echo(u'\r\n' + term.underline(u''.ljust(term.width - 1)))
    echo(u'\r\n\r\n')
    session.activity = 'Constructing a %s message' % (
        u'public' if u'public' in msg.tags else u'private',)
    return


def prompt_recipient(msg):
    """ Prompt for recipient of message. """
    # pylint: disable=R0914
    #         Too many local variables
    from x84.bbs import getterminal, LineEditor, echo, ini, list_users
    from x84.bbs import Selector
    import difflib
    term = getterminal()
    echo(u"ENtER %s, OR '%s' tO AddRESS All. %s to exit" % (
        term.bold_yellow(u'hANdlE'),
        term.bold_yellow(u'None'),
        term.bold_yellow_underline('Escape'),))
    echo(u'\r\n\r\n')
    max_user = ini.CFG.getint('nua', 'max_user')
    lne = LineEditor(max_user, msg.recipient or u'None')
    lne.highlight = term.yellow_reverse
    echo(term.clear_eol + u'   RECiPiENt: ')
    recipient = lne.read()
    if recipient is None or lne.quit:
        return False
    userlist = list_users()
    if recipient in userlist:
        msg.recipient = recipient
        return True
    elif len(recipient) != 0 and recipient != 'None':
        for match in difflib.get_close_matches(recipient, userlist):
            blurb = u'did YOU MEAN: %s ?' % (match,)
            inp = Selector(yloc=term.height - 1,
                           xloc=term.width - 22,
                           width=20, left=u'YES', right=u'NO')
            echo(u''.join((
                u'\r\n',
                term.move(inp.yloc, inp.xloc - len(blurb)),
                term.clear_eol,
                term.bold_yellow(blurb))))
            selection = inp.read()
            echo(term.move(inp.yloc, 0) + term.clear_eol)
            if selection == u'YES':
                msg.recipient = match
                return True
            if selection is None or inp.quit:
                return False
    else:
        blurb = u' NO RECiPiENT; POSt tO PUbliC? '
        inp = Selector(yloc=term.height - 1,
                       xloc=term.width - 22,
                       width=20, left=u'YES', right=u'NO')
        echo(u''.join((
            u'\r\n',
            term.move(inp.yloc, inp.xloc - len(blurb)),
            term.clear_eol,
            term.bold_yellow(blurb))))
        selection = inp.read()
        echo(term.move(inp.yloc, 0) + term.clear_eol)
        if selection == u'YES':
            msg.recipient = None
            return True


def prompt_subject(msg):
    """ Prompt for subject of message. """
    from x84.bbs import getterminal, LineEditor, echo, ini
    term = getterminal()
    max_subject = int(ini.CFG.getint('msg', 'max_subject'))
    lne = LineEditor(max_subject, msg.subject)
    lne.highlight = term.yellow_reverse
    echo(u'\r\n\r\n     SUBjECt: ')
    subject = lne.read()
    if subject is None or 0 == len(subject):
        return False
    msg.subject = subject
    return True


def prompt_tags(msg):
    """ Prompt for and return tags wished for message. """
    # pylint: disable=R0914,W0603
    #         Too many local variables
    #         Using the global statement
    from x84.bbs import DBProxy, echo, getterminal, getsession
    from x84.bbs import Ansi, LineEditor
    session, term = getsession(), getterminal()
    tagdb = DBProxy('tags')
    msg_onlymods = (u"\r\nONlY MEMbERS Of thE '%s' OR '%s' "
                    "GROUP MAY CREAtE NEW tAGS." % (
                        term.bold_yellow('sysop'),
                        term.bold_blue('moderator'),))
    msg_invalidtag = u"\r\n'%s' is not a valid tag."
    prompt_tags1 = u"ENtER %s, COMMA-dEliMitEd. " % (
        term.bold_red('TAG(s)'),)
    prompt_tags2 = u"OR '/list', %s:quit\r\n : " % (
        term.bold_yellow_underline('Escape'),)
    while True:
        # Accept user input for multiple 'tag's, or /list command
        echo(u'\r\n\r\n')
        echo(prompt_tags1)
        echo(prompt_tags2)
        width = term.width - 6
        sel_tags = u', '.join(msg.tags)
        inp_tags = LineEditor(width, sel_tags).read()
        if inp_tags is not None and 0 == len(inp_tags.strip()):
            # no tags must be (private ..)
            msg.tags = set()
            return True
        if inp_tags is None or inp_tags.strip().lower() == '/quit':
            return False
        elif inp_tags.strip().lower() == '/list':
            # list all available tags, and number of messages
            echo(u'\r\n\r\nTags: \r\n')
            all_tags = sorted(tagdb.items())
            if 0 == len(all_tags):
                echo(u'None !'.center(term.width / 2))
            else:
                echo(Ansi(u', '.join(([u'%s(%d)' % (_key, len(_value),)
                                       for (_key, _value) in all_tags]))
                          ).wrap(term.width - 2))
            continue
        echo(u'\r\n')

        # search input as valid tag(s)
        tags = set([inp.strip().lower() for inp in inp_tags.split(',')])
        err = False
        for tag in tags.copy():
            if not tag in tagdb and not (
                    'sysop' in session.user.groups or
                    'moderator' in session.user.groups):
                tags.remove(tag)
                echo(msg_invalidtag % (term.bold_red(tag),))
                err = True
        if err:
            echo(msg_onlymods)
            continue
        msg.tags = tags
        return True


def prompt_public(msg):
    """ Prompt for/enforce 'public' tag of message for unaddressed messages.
    """
    from x84.bbs import getterminal, echo, Selector
    term = getterminal()

    if 'public' in msg.tags:
        # msg is addressed to None, and is tagged as 'public',
        return True # quit/escape

    if msg.recipient is None:
        # msg is addressed to nobody, force tag as 'public' or cancel,
        blurb = u"POStS AddRESSEd tO 'None' MUSt bE PUbliC!"
        inp = Selector(yloc=term.height - 1,
                       xloc=term.width - 22,
                       width=20,
                       left=u'Ok', right=u'CANCEl')
        echo(term.move(inp.yloc, inp.xloc - len(blurb)) + term.clear_eol)
        echo(term.bold_red(blurb))
        selection = inp.read()
        echo(term.move(inp.yloc, 0) + term.clear_eol)
        if selection == u'Ok':
            msg.tags.add(u'public')
            return True
        return False # quit/escape

    # not specified; you don't want this msg public? confirm,
    inp = Selector(yloc=term.height - 1,
                   xloc=term.width - 24,
                   width=20,
                   left=u'YES!', right=u'NO')
    blurb = u'SENd PRiVAtE POSt?'
    echo(term.move(inp.yloc, inp.xloc - len(blurb)))
    echo(term.bold_yellow(blurb))
    selection = inp.read()
    echo(term.move(inp.yloc, 0) + term.clear_eol)
    if selection == u'NO':
        msg.tags.add(u'public')
        return True
    elif selection == u'YES!':
        if u'public' in msg.tags:
            msg.tags.remove(u'public')
        return True
    return False # quit/escape


def prompt_body(msg):
    """ Prompt for 'body' of message, executing 'editor' script. """
    from x84.bbs import echo, Selector, getterminal, getsession, gosub
    term = getterminal()
    session = getsession()
    inp = Selector(yloc=term.height - 1,
                   xloc=term.width - 22,
                   width=20,
                   left=u'CONtiNUE', right=u'CANCEl')
    blurb = u'CONtiNUE tO Edit MESSAGE bOdY'
    echo(u'\r\n\r\n')
    echo(term.move(inp.yloc, inp.xloc - len(blurb)))
    echo(term.bold_yellow(blurb))
    selection = inp.read()
    echo(term.move(inp.yloc, 0) + term.clear_eol)
    if selection != u'CONtiNUE':
        return False
    if 0 != len(session.user.get('draft', u'')):
        inp = Selector(yloc=term.height - 1,
                       xloc=term.width - 22,
                       width=20,
                       left=u'REStORE', right=u'ERASE')
        blurb = u'CONtiNUE PREViOUSlY SAVEd dRAft ?'
        echo(u'\r\n\r\n')
        echo(term.move(inp.yloc, inp.xloc - len(blurb)))
        echo(term.bold_yellow(blurb))
        selection = inp.read()
        echo(term.move(inp.yloc, 0) + term.clear_eol)
        if selection == u'REStORE':
            msg.body = session.user['draft']
    echo(u'\r\n\r\n')
    session.user['draft'] = msg.body
    if gosub('editor', 'draft'):
        echo(u'\r\n\r\n' + term.normal)
        msg.body = session.user.get('draft', u'')
        del session.user['draft']
        return 0 != len(msg.body.strip())
    return False


def prompt_send():
    """ Prompt for continue/cancel """
    from x84.bbs import echo, Selector, getterminal
    term = getterminal()
    inp = Selector(yloc=term.height - 1,
                   xloc=term.width - 22,
                   width=20,
                   left=u'CONtiNUE', right=u'CANCEl')
    blurb = u'CONtiNUE tO SENd MESSAGE'
    echo(term.move(inp.yloc, inp.xloc - len(blurb)))
    echo(term.bold_yellow(blurb))
    selection = inp.read()
    echo(term.move(inp.yloc, 0) + term.clear_eol)
    if selection != u'CONtiNUE':
        return False
    return True


def prompt_abort():
    """ Prompt for continue/abort """
    from x84.bbs import echo, Selector, getterminal
    term = getterminal()
    inp = Selector(yloc=term.height - 1,
                   xloc=term.width - 22,
                   width=20,
                   left=u'CONtiNUE', right=u'AbORt')
    blurb = u'CONtiNUE MESSAGE?'
    echo(u'\r\n\r\n')
    echo(term.move(inp.yloc, inp.xloc - len(blurb)))
    echo(term.bold_yellow(blurb))
    selection = inp.read()
    echo(term.move(inp.yloc, 0) + term.clear_eol)
    if selection == u'AbORt':
        return True
    return False


def main(msg=None):
    """ Main procedure. """
    from x84.bbs import Msg, getsession
    session = getsession()
    if msg is None:
        msg = Msg()
        msg.tags = ('public',)
    banner()
    while True:
        session.activity = 'Constructing a %s message' % (
            u'public' if u'public' in msg.tags else u'private',)
        if not prompt_recipient(msg):
            break
        if not prompt_subject(msg):
            break
        if not prompt_tags(msg):
            break
        if not prompt_public(msg):
            break
        if not prompt_body(msg):
            break
        display_msg(msg)
        if not prompt_send():
            break
        msg.save()
        return True
    return False
